import NDATools
from NDATools.Configuration import *
from NDATools.Utils import *
from NDATools.Validation import Validation
from NDATools.BuildPackage import SubmissionPackage
from NDATools.Submission import Submission
import argparse
import logging
import os
import pkg_resources


def parse_args():
    parser = argparse.ArgumentParser(
        description='This application allows you to validate files and submit data into NDA. '
                    'You must enter a list of at least one file to be validated. '
                    'If your data contains manifest files, you must specify the location of the manifests. '
                    'If your data also includes associated files, you must enter a list of at least one directory '
                    'where the associated files are saved. Alternatively, if any of your data is stored in AWS, you must'
                    ' provide your account credentials, the AWS bucket, and a prefix, if it exists.  '
                    'Any files that are created while running the client (ie. results files) will be downloaded in '
                    'your home directory under NDAValidationResults. If your submission was interrupted in the middle'
                    ', you may resume your upload by entering a valid submission ID. ',
        usage='%(prog)s <file_list>')

    parser.add_argument('files', metavar='<file_list>', type=str, nargs='+', action='store',
                        help='Returns validation results for list of files')

    parser.add_argument('-l', '--listDir', metavar='<directory_list>', type=str, nargs='+', action='store',
                        help='Specifies the directories in which the associated files are files located.')

    parser.add_argument('-m', '--manifestPath', metavar='<arg>', type=str, nargs='+', action='store',
                        help='Specifies the directories in which the manifest files are located')

    parser.add_argument('-s3', '--s3Bucket', metavar='<arg>', type=str, action='store',
                        help='Specifies the s3 bucket in which the associated files are files located.')

    parser.add_argument('-pre', '--s3Prefix', metavar='<arg>', type=str, action='store',
                        help='Specifies the s3 prefix in which the associated files are files located.')

    parser.add_argument('-w', '--warning', action='store_true',
                        help='Returns validation warnings for list of files')

    parser.add_argument('-a', '--alternateEndpoint', metavar='<arg>', type=str, action='store',
                        help='An alternate upload location for the submission package')

    parser.add_argument('-b', '--buildPackage', action='store_true',
                        help='Flag whether to construct the submission package')

    parser.add_argument('-c', '--collectionID', metavar='<arg>', type=int, action='store',
                        help='The NDA collection ID')

    parser.add_argument('-d', '--description', metavar='<arg>', type=str, nargs='+', action='store',
                        help='The description of the submission')

    parser.add_argument('-t', '--title', metavar='<arg>', type=str, nargs='+', action='store',
                        help='The title of the submission')

    parser.add_argument('-u', '--username', metavar='<arg>', type=str, action='store',
                        help='NDA username')

    parser.add_argument('-ak', '--accessKey', metavar='<arg>', type=str, action='store',
                        help='AWS access key')

    parser.add_argument('-sk', '--secretKey', metavar='<arg>', type=str, action='store',
                        help='AWS secret key')

    parser.add_argument('-p', '--password', metavar='<arg>', type=str, action='store',
                        help='NDA password')

    parser.add_argument('-s', '--scope', metavar='<arg>', type=str, action='store',
                        help='Flag whether to validate using a custom scope. Must enter a custom scope')

    parser.add_argument('-r', '--resume', action='store_true',
                        help='Restart an in-progress submission, resuming from the last successful part in a multi-part'
                             'upload. Must enter a valid submission ID.')

    parser.add_argument('-v', '--validationAPI', metavar='<arg>', type=str, action='store',
                        help='URL of the validation tool API')

    parser.add_argument('-j', '--JSON', action='store_true',
                        help='Flag whether to additionally download validation results in JSON format.')

    parser.add_argument('-wt', '--workerThreads', metavar='<arg>', type=int, action='store',
                        help='Number of worker threads')

    parser.add_argument('-bc', '--batch', metavar='<arg>', type=int, action='store',
                        help='Batch size')

    parser.add_argument('--hideProgress', action='store_true', help='Hides upload/processing progress')

    parser.add_argument('--skipLocalAssocFileCheck', action='store_true', help='Not recommended UNLESS you have already'
                        ' verified all paths for associated data files are correct')

    args = parser.parse_args()

    return args


def configure(args):
    # create a new config file in user's home directory if one does not exist

    NDATools.Utils.logging.getLogger().setLevel(logging.DEBUG)
    if os.path.isfile(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg')):
        config = ClientConfiguration(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg'), args.username,
                                     args.password, args.accessKey, args.secretKey)
    else:
        config = ClientConfiguration('clientscripts/config/settings.cfg', args.username, args.password, args.accessKey,
                                     args.secretKey)
        config.read_user_credentials()
        config.make_config()

    if args.collectionID:
        config.collection_id = args.collectionID
    if args.alternateEndpoint:
        config.endpoint_title = args.alternateEndpoint
    if args.listDir:
        config.directory_list = args.listDir
    if args.manifestPath:
        config.manifest_path = args.manifestPath
    if args.s3Bucket:
        config.source_bucket = args.s3Bucket
    if args.s3Prefix:
        config.source_prefix = args.s3Prefix
    if args.title:
        config.title = ' '.join(args.title)
    if args.description:
        config.description = ' '.join(args.description)
    if args.scope:
        config.scope = args.scope[0]
    if args.validationAPI:
        config.validation_api = args.validationAPI[0]
    if args.JSON:
        config.JSON = True
    config.hideProgress = args.hideProgress
    if args.skipLocalAssocFileCheck:
        config.skip_local_file_check = True

    return config


class Status:
    UPLOADING = 'Uploading'
    SYSERROR = 'SystemError'


def resume_submission(submission_id, batch, config=None):
    submission = Submission(id=submission_id, full_file_path=None, config=config, resume=True, batch_size=batch)
    submission.check_status()
    if submission.status == Status.UPLOADING:
        directories = config.directory_list
        source_bucket = config.source_bucket
        source_prefix = config.source_prefix

        if submission.incomplete_files and submission.found_all_files(directories, source_bucket, source_prefix,
                                                                       retry_allowed=True):
            # if not config.skip_local_file_check:
            submission.check_submitted_files()
            submission.complete_partial_uploads()
            submission.submission_upload(hide_progress=config.hideProgress)
        else:
           submission.submission_upload(hide_progress=config.hideProgress)

    else:
        print('Submission Completed with status {}'.format(submission.status))
        return


def validate_files(file_list, warnings, build_package, threads, config=None):

    validation = Validation(file_list, config=config, hide_progress=config.hideProgress, thread_num=threads, allow_exit=True)
    print('\nValidating files...')
    validation.validate()

    for (response, file) in validation.responses:
        if response['status'] == Status.SYSERROR:
            print('\nSystemError while validating: {}'.format(file))
            print('Please contact NDAHelp@mail.nih.gov')
        elif response['errors'] != {}:
            print('\nError! Check file: {}'.format(file))
    validation.output()
    print('Validation report output to: {}'.format(validation.log_file))

    if warnings:
        validation.get_warnings()
        print('Warnings output to: {}'.format(validation.log_file))

    else:
        if validation.w:
            print('\nNote: Your data has warnings. To save warnings, run again with -w argument.')
    print('\nAll files have finished validating.')

    # Test if no files passed validation, exit
    if not any(map(lambda x: not validation.uuid_dict[x]['errors'], validation.uuid_dict)):
        print('No files passed validation, please correct any errors and validate again.')
        return
    # If some files passed validation, show files with and without errors
    else:
        print('\nThe following files passed validation:')
        for uuid in validation.uuid_dict:
            if not validation.uuid_dict[uuid]['errors']:
                print('UUID {}: {}'.format(uuid, validation.uuid_dict[uuid]['file']))
        if validation.e:
            print('\nThese files contain errors:')
            for uuid in validation.uuid_dict:
                if validation.uuid_dict[uuid]['errors']:
                    print('UUID {}: {}'.format(uuid, validation.uuid_dict[uuid]['file']))
    # If some files had errors, give option to submit just the files that passed
    if validation.e and build_package:
        while True:
            proceed = input('Some files have errors, do you want to continue '
                            'and submit ONLY the files that have passed validation? <Yes/No>: ')
            if str(proceed).lower() == 'no':
                return
            elif str(proceed).lower() == 'yes':
                validation.uuid = validation.verify_uuid()
                break
            else:
                print('Your answer <{}> was not recognized, please enter yes or no.'.format(str(proceed)))
                continue
    return([validation.uuid, validation.associated_files])


def build_package(uuid, associated_files, config):
    if not config.title:
        config.title = input('Enter title for dataset name:')
    if not config.description:
        config.description = input('Enter description for the dataset submission:')

    package = SubmissionPackage(uuid, associated_files, config=config, allow_exit=True)
    package.set_upload_destination(hide_input=False)
    directories = config.directory_list
    source_bucket = config.source_bucket
    source_prefix = config.source_prefix
    if associated_files:
        print('\nSearching for associated files...')
        package.file_search(directories, source_bucket, source_prefix, retry_allowed=True)
    print('Building Package')
    package.build_package()
    print('\n\nPackage Information:')
    print('validation results: {}'.format(package.validation_results))
    print('submission_package_uuid: {}'.format(package.submission_package_uuid))
    print('created date: {}'.format(package.create_date))
    print('expiration date: {}'.format(package.expiration_date))
    print('\nPackage finished building.\n')

    print('Downloading submission package.')
    package.download_package(hide_progress=config.hideProgress)
    print('\nA copy of your submission package has been saved to: {}'.
          format(os.path.join(package.package_folder, package.config.submission_packages)))

    return[package.package_id, package.full_file_path]


def submit_package(package_id, full_file_path, associated_files, threads, batch, config=None):
    submission = Submission(id=package_id, full_file_path=full_file_path, thread_num=threads, batch_size=batch, allow_exit=True, config=config)
    print('Requesting submission for package: {}'.format(submission.package_id))
    submission.submit()
    # see commit comment for commit #d2f4dad
    # we need to trigger the GET /id endpoint to move the submission status to complete if necessary
    submission.check_status()

    if submission.submission_id:
            print('Submission ID: {}'.format(str(submission.submission_id)))
        if associated_files:
            print('Preparing to upload associated files.')
            submission.submission_upload(hide_progress=config.hideProgress)
        if submission.status != Status.UPLOADING:
            print('\nYou have successfully completed uploading files for submission {} with status: {}'.format
                  (submission.submission_id, submission.status))
                   #do we want to include this??


def main():
    args = parse_args()
    config = configure(args)

    if args.resume:
        submission_id = args.files[0]
        resume_submission(submission_id, batch=args.batch, config=config)
    else:
        w = False
        bp = False
        if args.warning:
            w = True
        if args.buildPackage:
            bp = True
        validation_results = validate_files(args.files, w, bp, threads=args.workerThreads, config=config)
        if validation_results is not None:
            uuid = validation_results[0]
            associated_files = validation_results[1]
            # If user requested to build a package
            if bp:
                package_results = build_package(uuid, associated_files, config=config)
                package_id = package_results[0]
                full_file_path = package_results[1]
                submit_package(package_id=package_id, full_file_path=full_file_path, associated_files=associated_files,
                               threads=args.workerThreads, batch=args.batch, config=config)


if __name__ == "__main__":
    main()



from __future__ import with_statement
from __future__ import absolute_import
import sys
import getpass
if sys.version_info[0] < 3:
    import ConfigParser as configparser
    input = raw_input
    import thread
else:
    import configparser
    import _thread as thread
import os
from pkg_resources import resource_filename


class ClientConfiguration:
    def __init__(self, settings_file, username=None, password=None, access_key=None, secret_key=None):
        self.config = configparser.ConfigParser()
        if settings_file == os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg'):
            config_location = settings_file
        else:
            config_location = resource_filename(__name__, settings_file)
        self.config.read(config_location)

        self.validation_api = self.config.get("Endpoints", "validation")
        self.submission_package_api = self.config.get("Endpoints", "submission_package")
        self.submission_api = self.config.get("Endpoints", "submission")
        self.validationtool_api = self.config.get("Endpoints", "validationtool")
        self.datamanager_api = self.config.get("Endpoints", "data_manager")
        self.validation_results = self.config.get("Files", "validation_results")
        self.submission_packages = self.config.get("Files", "submission_packages")
        self.aws_access_key = self.config.get("User", "access_key")
        self.aws_secret_key = self.config.get("User", "secret_key")
        self.aws_session_token = ""
        if not self.config.has_option("User", "session_token"):
            fixed_config = configparser.RawConfigParser()
            fixed_config.read(config_location)
            fixed_config.set("User", "session_token", "")
            with open(config_location, 'w') as configfile:
                fixed_config.write(configfile)
        else:
            self.aws_session_token = self.config.get("User", "session_token")
        self.username = self.config.get("User", "username")
        self.password = self.config.get("User", "password")

        if username:
            self.username = username
        if password:
            self.password = password
        if access_key:
            self.aws_access_key = access_key
        if secret_key:
            self.aws_secret_key = secret_key
        self.collection_id = None
        self.endpoint_title = None
        self.scope = None
        self.directory_list = None
        self.manifest_path = None
        self.source_bucket = None
        self.source_prefix = None
        self.title = None
        self.description = None
        self.JSON = False
        self.hideProgress = False
        self.skip_local_file_check = False

    def make_config(self):
        file_path = os.path.join(os.path.expanduser('~'), '.NDATools')
        if not os.path.exists(file_path):
            os.makedirs(file_path)
        config_path = os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg')

        copy_config = configparser.ConfigParser()

        copy_config.add_section("Endpoints")
        copy_config.set("Endpoints", "data_manager", self.datamanager_api)
        copy_config.set("Endpoints", "validation", self.validation_api)
        copy_config.set("Endpoints", "submission_package", self.submission_package_api)
        copy_config.set("Endpoints", "submission", self.submission_api)
        copy_config.set("Endpoints", "validationtool", self.validationtool_api)

        copy_config.add_section("Files")
        copy_config.set("Files", "validation_results", self.validation_results)
        copy_config.set("Files", "submission_packages", self.submission_packages)

        copy_config.add_section("User")
        copy_config.set("User", "username", self.username)
        copy_config.set("User", "password", self.password)
        copy_config.set("User", "access_key", self.aws_access_key)
        copy_config.set("User", "secret_key", self.aws_secret_key)
        copy_config.set("User", "session_token", self.aws_session_token)

        with open(config_path, 'w') as configfile:
            copy_config.write(configfile)

    def read_user_credentials(self):
        if not self.username:
            self.username = input('Enter your NIMH Data Archives username:')
        if not self.password:
            self.password = getpass.getpass('Enter your NIMH Data Archives password:')

        if not self.aws_access_key:
            self.aws_access_key = getpass.getpass('Enter your aws_access_key. If none, hit "Enter:"')

        if not self.aws_secret_key:
            self.aws_secret_key = getpass.getpass('Enter your aws_secret_key. If none, hit "Enter":')



import NDATools
import requests
import json
import sys

__version__ = '0.2.1'
pypi_version = None
version_checked = False

def check_version():
    try:
        from packaging.version import parse
    except ImportError:
        from pip._vendor.packaging.version import parse
    # use https://test.pypi.org/pypi/{package}/json on test/release branches, use https://pypi.org on master
    print('Running NDATools Version {}'.format(NDATools.__version__))
    if parse(NDATools.__version__).is_devrelease:
        return
    url_pattern = 'https://pypi.org/pypi/{package}/json'
    package = 'nda-tools'
    """Return version of package on pypi.python.org using json."""
    req = requests.get(url_pattern.format(package=package))
    version = parse('0')
    if req.status_code == requests.codes.ok:
        j = json.loads(req.text)
        releases = j.get('releases', [])
        for release in releases:
            ver = parse(release)
            if not ver.is_prerelease:
                version = max(version, ver)
    NDATools.pypi_version = str(version)

    if NDATools.__version__ != NDATools.pypi_version:
        print("Your version of nda-tools is out of date. Please install the latest version ({}) from PyPi or GitHub and "
              "try again.".format(NDATools.pypi_version))

    NDATools.version_checked = True


if not NDATools.version_checked:
    check_version()




from __future__ import with_statement
from __future__ import absolute_import
import re
import signal
import sys
import getpass
import time
import threading
import requests
from requests.adapters import HTTPAdapter
import requests.packages.urllib3.util
import json
import signal
import os
import logging

from NDATools.Configuration import ClientConfiguration

if os.path.isfile(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg')):
    config = ClientConfiguration(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg'))
else:
    config = ClientConfiguration('clientscripts/config/settings.cfg')
validation_results_dir = os.path.join(os.path.expanduser('~'), config.validation_results)
if not os.path.exists(validation_results_dir):
    os.mkdir(validation_results_dir)
log_file = os.path.join(validation_results_dir, "debug_log_{}.txt").format(time.strftime("%Y%m%dT%H%M%S"))
print('Opening log: {}'.format(log_file))
logging.basicConfig(filename=log_file, level=logging.INFO, format="%(asctime)s:%(levelname)s:%(message)s")

if sys.version_info[0] < 3:
    import ConfigParser as configparser

    input = raw_input
    import thread
else:
    import configparser
    import _thread as thread

import xml.etree.ElementTree as ET

class Protocol(object):
    CSV = "csv"
    XML = "xml"
    JSON = "json"

    @staticmethod
    def get_protocol(cls):
        return cls.JSON

def api_request(api, verb, endpoint, data=None, session=None):

    retry = requests.packages.urllib3.util.retry.Retry(
        total=20,
        read=20,
        connect=20,
        backoff_factor=3,
        status_forcelist=(400, 403, 404, 500, 502, 504)
    )

    headers = {'accept': 'application/json'}
    auth = None
    if isinstance(api, Protocol):
        if api.get_protocol(api) == Protocol.CSV:
            headers.update({'content-type': 'text/csv'})
        elif api.get_protocol(api) == Protocol.XML:
            headers.update({'content-type': 'text/xml'})
    else:
        auth = requests.auth.HTTPBasicAuth(api.config.username, api.config.password)
        headers.update({'content-type': 'application/json'})

    if not session:
        session = requests.session()
        session.mount(endpoint, HTTPAdapter(max_retries=retry))
    r = None
    response = None
    try:
        r = session.send(requests.Request(verb, endpoint, headers, auth=auth, data=data).prepare(),
                         timeout=300, stream=False)

    except requests.exceptions.RequestException as e:
        print('\nAn error occurred while making {} request, check your endpoint configuration:\n'.
              format(e.request.method))
        logging.error(e)
        if api.__class__.__name__.endswith('Task'):
            api.shutdown_flag.set()
            thread.interrupt_main()
        exit_client(signal.SIGTERM)

    if r and r.ok:
        if api.__class__.__name__ == 'Download':
            return r, session
        else:
            try:
                response = json.loads(r.text)
            except ValueError:
                logging.error(ValueError)
                print('Your request returned an unexpected response, please check your endpoints.\n'
                      'Action: {}\n'
                      'Endpoint:{}\n'
                      'Status:{}\n'
                      'Reason:{}'.format(verb, endpoint, r.status_code, r.reason))
                if api.__class__.__name__.endswith('Task'):
                    api.shutdown_flag.set()
                    thread.interrupt_main()
                else:
                    raise Exception(ValueError)

    if r.status_code == 401:
        m = 'The NDA username or password is not recognized.'
        print(m)
        logging.error(m)
        r.raise_for_status()

    elif r.status_code == 400:
        response = json.loads(r.text)
        m = response['error'] + ': ' + response['message']
        print(m)
        logging.error(m)
        r.raise_for_status()

    elif r.status_code in (500, 502, 503, 504):
        response = r.text
        m = response
        print(m)
        logging.error(m)
        r.raise_for_status()

    return response, session


def exit_client(signal, frame=None, message=None):
    for t in threading.enumerate():
        try:
            t.shutdown_flag.set()
        except AttributeError as e:
            continue
    if message:
        print('\n\n{}'.format(message))
    else:
        print('\n\nExit signal received, shutting down...')
    print('Please contact NDAHelp@mail.nih.gov if you need assistance.')
    sys.exit(1)


def parse_local_files(directory_list, no_match, full_file_path, no_read_access, skip_local_file_check):
    """
    Iterates through associated files generate a dictionary of full filepaths and file sizes.

    :param directory_list: List of directories
    :param no_match: Stores list of invalid paths
    :param full_file_path: Dictionary of tuples that store full filepath and file size
    :param no_read_access: List of files that user does not have access to
    :return: Modifies references to no_match, full_file_path, no_read_access
    """
    files_to_match = len(no_match)
    logging.debug('Starting local directory search for {} files'.format(str(files_to_match)))
    progress_counter = int(files_to_match*0.05)
    for file in no_match[:]:
        if progress_counter == 0:
            logging.debug('Found {} files out of {}'.format(str(files_to_match - len(no_match)), str(files_to_match)))
            progress_counter = int(files_to_match*0.05)
        file_key = sanitize_file_path(file)
        for d in directory_list:
            if skip_local_file_check:
                file_name = os.path.join(d, file)
                try:
                    full_file_path[file_key] = (file_name, os.path.getsize(file_name))
                    no_match.remove(file)
                    progress_counter -= 1
                except (OSError, IOError) as err:
                    if err.errno == 13:
                        print('Permission Denied: {}'.format(file_name))
                    continue
                break
            else:
                if os.path.isfile(file):
                    file_name = file
                elif os.path.isfile(os.path.join(d, file)):
                    file_name = os.path.join(d, file)
                else:
                    continue
                if not check_read_permissions(file_name):
                    no_read_access.add(file_name)
                full_file_path[file_key] = (file_name, os.path.getsize(file_name))
                no_match.remove(file)
                break
    logging.debug('Local directory search complete, found {} files out of {}'.format(str(files_to_match - len(no_match)), str(files_to_match)))


def sanitize_file_path(file):
    """
    Replaces backslashes with forward slashes and removes leading / or drive:/.

    :param file: Relative or absolute filepath
    :return: Relative or absolute filepath with leading / or drive:/ removed
    """
    # Sanitize all backslashes (\) with forward slashes (/)
    file_key = file.replace('\\', '/').replace('//', '/')
    # If Mac/Linux full path
    if re.search(r'^/.+$', file):
        file_key = file.split('/', 1)[1]
    # If Windows full path
    elif re.search(r'^\D:/.+$', file_key):
        file_key = file_key.split(':/', 1)[1]
    return file_key


def check_read_permissions(file):
    try:
        open(file)
        return True
    except (OSError, IOError) as err:
        if err.errno == 13:
            print('Permission Denied: {}'.format(file))
    return False



from NDATools.DataManager import DataManager
import boto3


class S3Authentication:

    def __init__(self, config):
        self.config = config
        self.aws_session_token = None
        if config.aws_secret_key == "" and config.aws_access_key == "":
            data_manager_credentials = DataManager(config).get_data_manager_credentials
            self.aws_access_key = data_manager_credentials.get('aws_access_key_id')
            self.aws_secret_key =  data_manager_credentials.get('aws_secret_access_key')
            self.aws_session_token = data_manager_credentials.get('aws_session_token')
        else:
            self.aws_access_key = config.aws_access_key
            self.aws_secret_key = config.aws_secret_key
            if config.aws_session_token != "":
                self.aws_session_token = config.aws_session_token
        self.credentials = self.get_credentials

    @property
    def get_credentials(self):
        return {'aws_access_key_id': self.aws_access_key,
                'aws_secret_access_key': self.aws_secret_key,
                'aws_session_token': self.aws_session_token}

    def get_s3_client(self):
        return boto3.session.Session(**self.credentials).client('s3')

    @staticmethod
    def get_s3_client_with_config(aws_access_key, aws_secret_key, aws_session_token):
        return boto3.session.Session(aws_access_key_id=aws_access_key,
                                     aws_secret_access_key=aws_secret_key,
                                     aws_session_token=aws_session_token,
                                     region_name='us-east-1').client('s3')

    def get_s3_resource(self, s3_config):
        return boto3.Session(**self.credentials).resource('s3', config=s3_config)



from __future__ import with_statement
from __future__ import absolute_import
import sys
import csv
import multiprocessing
from tqdm import tqdm
if sys.version_info[0] < 3:
    import Queue as queue
    input = raw_input
else:
    import queue
from NDATools.Configuration import *
from NDATools.Utils import *
import threading
import signal


class Validation:
    def __init__(self, file_list, config, hide_progress, allow_exit=False, thread_num=None):
        self.config = config
        self.hide_progress = hide_progress
        self.api = self.config.validation_api.strip('/')
        self.thread_num = max([1, multiprocessing.cpu_count() - 1])
        if thread_num:
            self.thread_num = thread_num
        self.scope = self.config.scope
        self.api_scope = self.api
        if self.scope is not None:
            self.api_scope = "".join([self.api, '/?scope={}'.format(self.scope)])
        self.file_queue = queue.Queue()
        self.result_queue = queue.Queue()
        self.file_list = file_list
        self.directory_list = self.config.directory_list
        self.associated_files = []
        self.uuid = []
        self.uuid_dict = {}
        self.responses = []
        self.date = time.strftime("%Y%m%dT%H%M%S")
        self.e = False
        self.w = False
        self.manifest_data_files = []
        self.manifest_path = self.config.manifest_path
        self.validation_result_dir = os.path.join(os.path.expanduser('~'), self.config.validation_results)
        if not os.path.exists(self.validation_result_dir):
            os.mkdir(self.validation_result_dir)
        if self.config.JSON:
            self.log_file = os.path.join(self.validation_result_dir, 'validation_results_{}.json'.format(self.date))
        else:
            self.log_file = os.path.join(self.validation_result_dir, 'validation_results_{}.csv'.format(self.date))

        self.field_names = ['FILE', 'ID', 'STATUS', 'EXPIRATION_DATE', 'ERRORS', 'COLUMN', 'MESSAGE', 'RECORD']
        self.validation_progress = None
        self.exit = allow_exit

    """
    Validates a list of csv files and saves a new csv file with the results in the Home directory.
    If data includes associated files and user has not entered any directories where they are saved, user is prompted to 
    enter a list of directories only if they have also indicated to build a package for submission.
    """

    def validate(self):
        if not self.hide_progress:
            self.validation_progress = tqdm(total=len(self.file_list), position=0, ascii=os.name == 'nt')
        workers = []
        for x in range(self.thread_num):
            worker = Validation.ValidationTask(self.file_queue, self.result_queue, self.api, self.api_scope, self.responses, self.validation_progress, self.exit)
            workers.append(worker)
            worker.daemon = True
            worker.start()
        for file in self.file_list:
            self.file_queue.put(file)
        self.file_queue.put("STOP")
        while any(map(lambda x: x.is_alive(), workers)):
            time.sleep(0.1)
        if not self.hide_progress:
            self.validation_progress.close()
        while True:
            try:
                (response, file) = self.result_queue.get()
                self.result_queue.task_done()
                self.responses.append((response, file))
                if self.result_queue.empty():
                    break
            except queue.Empty:
                break
        for (response, file) in self.responses:
            if response['status'] == Status.SYSERROR:
                self.e = True
                error = " ".join([response['status'], response['errors']['system'][0]['message']])
                raise Exception(error)
            elif response['errors'] != {}:
                self.e = True
            if response['associated_file_paths'] and response['errors'] == {}:
                self.associated_files.append(response['associated_file_paths'])
            if response['warnings'] != {}:
                self.w = True
            if response['status'] == Status.PENDING_MANIFEST and response['errors'] == {}:
                self.process_manifests(response)
                response, session = api_request(self, "GET", "/".join([self.api, response['id']]))
                self.associated_files.append(response['associated_file_paths'])

            self.uuid.append(response['id'])
            self.uuid_dict[response['id']] = {'file': file, 'errors': response['errors'] != {}}

    def output(self):
        if self.config.JSON:
            json_data = dict(Results=[])
            for (response, file) in self.responses:
                file_name = self.uuid_dict[response['id']]['file']
                json_data['Results'].append({
                    'File': file_name,
                    'ID': response['id'],
                    'Status': response['status'],
                    'Expiration Date': response['expiration_date'],
                    'Errors': response['errors']
                })
            with open(self.log_file, 'w') as outfile:
                json.dump(json_data, outfile)
        else:
            if sys.version_info[0] < 3:
                csvfile = open(self.log_file, 'wb')
            else:
                csvfile = open(self.log_file, 'w', newline='')
            writer = csv.DictWriter(csvfile, fieldnames=self.field_names)
            writer.writeheader()
            for (response, file) in self.responses:
                file_name = self.uuid_dict[response['id']]['file']
                if response['errors'] == {}:
                    writer.writerow(
                        {'FILE': file_name, 'ID': response['id'], 'STATUS': response['status'],
                         'EXPIRATION_DATE': response['expiration_date'], 'ERRORS': 'None', 'COLUMN': 'None',
                         'MESSAGE':'None','RECORD': 'None'})
                else:
                    for error, value in response['errors'].items():
                        for v in value:
                            column = v['columnName']
                            message = v['message']
                            try:
                                record = v['recordNumber']
                            except KeyError:
                                record = ' '
                            writer.writerow(
                                {'FILE': file_name, 'ID': response['id'], 'STATUS': response['status'],
                                 'EXPIRATION_DATE': response['expiration_date'], 'ERRORS': error, 'COLUMN': column,
                                 'MESSAGE': message, 'RECORD': record})
            csvfile.close()

    def get_warnings(self):
        if self.config.JSON:
            json_data = dict(Results=[])
            for (response, file) in self.responses:
                file_name = self.uuid_dict[response['id']]['file']
                json_data['Results'].append({
                    'File': file_name,
                    'ID': response['id'],
                    'Status': response['status'],
                    'Expiration Date': response['expiration_date'],
                    'Warnings': response['warnings']
                })
                new_path = ''.join([self.validation_result_dir, '/validation_warnings_', self.date, '.json'])
                with open(new_path, 'w') as outfile:
                    json.dump(json_data, outfile)
        else:
            new_path = ''.join([self.validation_result_dir, '/validation_warnings_', self.date, '.csv'])
            if sys.version_info[0] < 3:
                csvfile = open(new_path, 'wb')
            else:
                csvfile = open(new_path, 'w', newline='')
            fieldnames = ['FILE', 'ID', 'STATUS', 'EXPIRATION_DATE', 'WARNINGS', 'MESSAGE', 'COUNT']
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            for (response, file) in self.responses:
                file_name = self.uuid_dict[response['id']]['file']
                if response['warnings'] == {}:
                    writer.writerow(
                        {'FILE': file_name, 'ID': response['id'], 'STATUS': response['status'],
                         'EXPIRATION_DATE': response['expiration_date'], 'WARNINGS': 'None', 'MESSAGE': 'None',
                         'COUNT': '0'})
                else:
                    for warning, value in response['warnings'].items():
                        m = {}
                        warning_list = []
                        for v in value:
                            count = 1
                            if v not in warning_list:
                                warning_list.append(v)
                                message = v['message']
                                m[message] = count
                            else:
                                count = m[message] + 1
                                m[message] = count
                        for x in m:
                            writer.writerow(
                                {'FILE': file_name, 'ID': response['id'], 'STATUS': response['status'],
                                 'EXPIRATION_DATE': response['expiration_date'], 'WARNINGS': warning, 'MESSAGE': x,
                                 'COUNT': m[x]})
            csvfile.close()

    def verify_uuid(self):
        uuid_list = []
        for uuid in self.uuid_dict:
            if not self.uuid_dict[uuid]['errors']:
                uuid_list.append(uuid)

        for response in self.manifest_data_files:
            if response['id'] not in uuid_list:
                self.manifest_data_files.remove(response)

        return uuid_list

    def process_manifests(self, r, validation_results = None, yes_manifest = None, bulk_upload=False):
        if not self.manifest_path:
            if not self.exit:
                error = 'Missing Manifest File: You must include the path to your manifests files'
                raise Exception(error)
            else:
                manifest_path_input = input("\nYour data contains manifest files. Please enter a list of "
                                            "the complete paths to the folder where your manifests are located,"
                                            "separated by a space:")
                self.manifest_path = manifest_path_input.split(' ')

        if not yes_manifest:
            yes_manifest = []

        no_manifest = set()

        if validation_results:
            self.validation_result = validation_results
        else:
            self.validation_result = Validation.ValidationManifestResult(r, self.hide_progress)
        if not bulk_upload:

            for validation_manifest in self.validation_result.manifests:
                for m in self.manifest_path:
                    if validation_manifest.local_file_name not in yes_manifest:
                        try:
                            manifest_path = os.path.join(m, validation_manifest.local_file_name)
                            validation_manifest.upload_manifest(open(manifest_path, 'rb'))
                            yes_manifest.append(validation_manifest.local_file_name)
                            if validation_manifest.local_file_name in no_manifest:
                                no_manifest.remove(validation_manifest.local_file_name)
                            break
                        except IOError:
                            no_manifest.add(validation_manifest.local_file_name)
                        except FileNotFoundError:
                            no_manifest.add(validation_manifest.local_file_name)

                        except json.decoder.JSONDecodeError as e:
                            error = 'JSON Error: There was an error in your json file: {}\nPlease review and try again: {}\n'.format\
                                (validation_manifest.local_file_name, e)
                            raise Exception(error)

        for file in no_manifest:
            print('Manifest file not found:',file)

        if no_manifest:
            print(
                '\nYou must make sure all manifest files listed in your validation file'
                ' are located in the specified directory. Please try again.')

            retry = input('Press the "Enter" key to specify location(s) for manifest files and try again:')
            self.manifest_path = retry.split(' ')
            self.process_manifests(r, yes_manifest=yes_manifest, validation_results = self.validation_result)

        while not self.validation_result.status.startswith(Status.COMPLETE):
            response, session = api_request(self, "GET", "/".join([self.api, r['id']]))
            self.validation_result = Validation.ValidationManifestResult(response, self.hide_progress)
            for m in self.validation_result.manifests:
                if m.status == Status.ERROR:
                    error = 'JSON Error: There was an error in your json file: {}\nPlease review and try again: {}\n'.format(m.local_file_name, m.errors[0])
                    raise Exception(error)

    class ValidationManifest:

        def __init__(self, _manifest, hide):
            self.config = ClientConfiguration(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg'))
            self.status = _manifest['status']
            self.local_file_name = _manifest['localFileName']
            self.manifestUuid = _manifest['manifestUuid']
            self.errors = _manifest['errors']
            self.url = _manifest['_links']['self']['href']
            self.associated_files = []
            self.hide_progress = hide

        def upload_manifest(self, _fp):
            manifest_object = json.load(_fp)
            api_request(self, "PUT", self.url, data=json.dumps(manifest_object))

    class ValidationManifestResult:

        def __init__(self, _validation_result, hide):
            self.done = _validation_result['done']
            self.id = _validation_result['id']
            self.status = _validation_result['status']
            self.errors = _validation_result['errors']
            self.short_name = _validation_result['short_name']
            self.scope = _validation_result['scope']
            manifests = []
            for _manifest in _validation_result['manifests']:
                manifests.append(Validation.ValidationManifest(_manifest, hide))
            self.manifests = manifests

    class ValidationTask(threading.Thread, Protocol):
        def __init__(self, file_queue, result_queue, api, scope, responses, validation_progress, exit):
            threading.Thread.__init__(self)
            self.file_queue = file_queue
            self.result_queue = result_queue
            self.api = api
            self.api_scope = scope
            self.responses = responses
            self.progress_bar = validation_progress
            self.shutdown_flag = threading.Event()
            self.exit = exit

        @staticmethod
        def get_protocol(cls):
            return cls.CSV

        def run(self):
            while True and not self.shutdown_flag.is_set():
                polling = 0
                file_name = self.file_queue.get()
                if file_name == "STOP":
                    self.file_queue.put("STOP")
                    self.shutdown_flag.set()
                    break
                try:
                    file = open(file_name, 'r')
                except IOError:
                    message = 'This file does not exist in current directory: {}'.format(file_name)
                    if self.progress_bar:
                        self.progress_bar.close()
                    if self.exit:
                        exit_client(signal=signal.SIGTERM, message=message)
                    else:
                        error = " ".join(['FileNotFound:', message])
                        raise Exception(error)

                data = file.read()

                response, session = api_request(self, "POST", self.api_scope, data)
                while response and not response['done']:
                    response, session = api_request(self, "GET", "/".join([self.api, response['id']]), session=session)
                    time.sleep(0.1)
                    polling += 1
                    if polling == 50:
                        polling = 0
                if response:
                    response, session = api_request(self, "GET", "/".join([self.api, response['id']]), session=session)
                    self.result_queue.put((response, file_name))
                    if self.progress_bar:
                        self.progress_bar.update(n=1)
                # Stop thread after adding validation response
                self.file_queue.task_done()


class Status:
    UPLOADING = 'Uploading'
    SYSERROR = 'SystemError'
    PENDING_MANIFEST = 'PendingManifestFiles'
    COMPLETE = 'Complete'
    ERROR = 'Error'



from __future__ import with_statement
from __future__ import absolute_import
import sys

from NDATools.S3Authentication import S3Authentication

IS_PY2 = sys.version_info < (3, 0)

if IS_PY2:
    from Queue import Queue
    from io import open
else:
    from queue import Queue
import csv
from threading import Thread
import boto3
import botocore
import datetime
from boto3.s3.transfer import S3Transfer
import multiprocessing

from NDATools.TokenGenerator import *
from NDATools.Utils import *



class Worker(Thread):
    """ Thread executing tasks from a given tasks queue """
    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.daemon = True
        self.start()

    def run(self):
        while True:
            func, args, kargs = self.tasks.get()
            try:
                func(*args, **kargs)
            except Exception as e:
                # An exception happened in this thread
                print(e)
            finally:
                # Mark this task as done, whether an exception happened or not
                self.tasks.task_done()


class ThreadPool:
    """ Pool of threads consuming tasks from a queue """
    def __init__(self, num_threads):
        self.tasks = Queue(num_threads)
        for _ in range(num_threads):
            Worker(self.tasks)

    def add_task(self, func, *args, **kargs):
        """ Add a task to the queue """
        self.tasks.put((func, args, kargs))

    def map(self, func, args_list):
        """ Add a list of tasks to the queue """
        for args in args_list:
            self.add_task(func, args)

    def wait_completion(self):
        """ Wait for completion of all the tasks in the queue """
        self.tasks.join()


class Download(Protocol):

    def __init__(self, directory, config=None, verbose=False):
        if config:
            self.config = config
        else:
            self.config = ClientConfiguration()
        self.url = self.config.datamanager_api
        self.username = config.username
        self.password = config.password
        self.directory = directory
        self.download_queue = Queue()
        self.path_list = set()
        self.local_file_names = {}
        self.access_key = None
        self.secret_key = None
        self.session = None
        self.associated_files = False
        self.dsList = []
        self.verbose = verbose

    @staticmethod
    def get_protocol(cls):
        return cls.XML

    def verbose_print(self, *args):
        if self.verbose:
            print(' '.join(list(args)))

    def useDataManager(self):
        """ Download package files (not associated files) """

        payload = ('<?xml version="1.0" ?>\n' +
                   '<S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">\n' +
                   '<S:Body> <ns3:QueryPackageFileElement\n' +
                   'xmlns:ns4="http://dataManagerService"\n' +
                   'xmlns:ns3="http://gov/nih/ndar/ws/datamanager/server/bean/jaxb"\n' +
                   'xmlns:ns2="http://dataManager/transfer/model">\n' +
                   '<packageId>' + self.package + '</packageId>\n' +
                   '<associated>true</associated>\n' +
                   '</ns3:QueryPackageFileElement>\n' +
                   '</S:Body>\n' +
                   '</S:Envelope>')

        response, session = api_request(self, "POST", self.url, data=payload)

        root = ET.fromstring(response.text)
        packageFiles = root.findall(".//queryPackageFiles")
        for element in packageFiles:
            associated = element.findall(".//isAssociated")
            path = element.findall(".//path")
            alias = element.findall(".//alias")
            for a in associated:
                if a.text == 'false':
                    for p in path:
                        file = 's3:/' + p.text
                        self.path_list.add(file)
                    for al in alias:
                        alias_path = al.text

                    self.local_file_names[file] = alias_path

        self.verbose_print('Downloading package files for package {}.'.format(self.package))

    def searchForDataStructure(self, resume, dir):
        """ Download associated files listed in data structures """

        all_paths = self.path_list
        self.path_list = set()

        for path in all_paths:
            if 'Package_{}'.format(self.package) in path:
                file = path.split('/')[-1]
                shortName = file.split('.')[0]
                try:
                    ddr = requests.request("GET", "https://ndar.nih.gov/api/datadictionary/v2/datastructure/{}".format(
                        shortName))
                    ddr.raise_for_status()
                    dataStructureFile = path.split('gpop/')[1]
                    dataStructureFile = os.path.join(self.directory, dataStructureFile)
                    self.dataStructure = dataStructureFile
                    self.useDataStructure()
                    self.get_tokens()
                    self.start_workers(resume, prev_directory=dir)
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code == 404:
                        continue

    def useDataStructure(self, data_structure):
        try:
            with open(self.dataStructure, 'r', encoding='utf-8') as tsv_file:
                tsv = csv.reader(tsv_file, delimiter="\t")
                if data_structure:
                    ds = next(tsv)[1].split("_")[0]
                for row in tsv:
                    for element in row:
                        if element.startswith('s3://'):
                            self.path_list.add(element)
                            if data_structure:
                                e= element.split('/')
                                path = "".join(e[4:])
                                path = "/".join([ds, path])
                                self.local_file_names[element] = path

        except IOError as e:
            self.verbose_print(
                '{} not found. Please enter the correct path to your file and try again.'.format(self.dataStructure))
            raise e

    def get_links(self, links, files, filters=None):

        if links == 'datastructure':
            self.dataStructure = files[0]
            self.useDataStructure(data_structure=True)

        elif links == 'text':
            self.dataStructure = files[0]
            self.useDataStructure(data_structure=False)

        elif links == 'package':
            self.package = files[0]
            self.useDataManager()

        else:
            self.path_list = files

    def check_time(self):
        now_time = datetime.datetime.now()
        if now_time >= self.refresh_time:
            self.get_tokens()

    def get_tokens(self):
        start_time = datetime.datetime.now()
        generator = NDATokenGenerator(self.url)
        self.token = generator.generate_token(self.username, self.password)
        self.refresh_time = start_time + datetime.timedelta(hours=23, minutes=00)
        self.access_key = self.token.access_key
        self.secret_key = self.token.secret_key
        self.session = self.token.session

    def download_path(self, path, resume, prev_directory):
        filename = path.split('/')
        self.filename = filename[3:]
        key = '/'.join(self.filename)
        bucket = filename[2]

        if self.local_file_names:
            dir = (self.local_file_names[path]).split('/')
            self.newdir = dir[:-1]
            alias = self.local_file_names[path]
        else:
            self.newdir = filename[3:-1]
            alias = key

        self.newdir = '/'.join(self.newdir)
        self.newdir = os.path.join(self.directory, self.newdir)
        local_filename = os.path.join(self.directory, alias)

        downloaded = False

        # check previous downloads
        if resume:
            prev_local_filename = os.path.join(prev_directory, key)
            if os.path.isfile(prev_local_filename):
                downloaded = True

        if not downloaded:
            try:
                os.makedirs(self.newdir)
            except OSError as e:
                pass

            # check tokens
            self.check_time()

            s3transfer = S3Transfer(S3Authentication.get_s3_client_with_config(self.access_key,
                                                                               self.secret_key,
                                                                               self.session))

            try:
                s3transfer.download_file(bucket, key, local_filename)
                self.verbose_print('downloaded: {}'.format(path))

            except botocore.exceptions.ClientError as e:
                # If a client error is thrown, then check that it was a 404 error.
                # If it was a 404 error, then the bucket does not exist.
                error_code = int(e.response['Error']['Code'])
                if error_code == 404:
                    message = 'This path is incorrect: {}. Please try again.'.format(path)
                    self.verbose_print(message)
                    raise Exception(e)

                if error_code == 403:
                    message = '\nThis is a private bucket. Please contact NDAR for help: {}'.format(path)
                    self.verbose_print(message)
                    raise Exception(e)

    def start_workers(self, resume, prev_directory, thread_num=None):
        def download(path):
            self.download_path(path, resume, prev_directory)

        # Instantiate a thread pool with i worker threads
        self.thread_num = max([1, multiprocessing.cpu_count() - 1])
        if thread_num:
            self.thread_num = thread_num

        pool = ThreadPool(self.thread_num)

        # Add the jobs in bulk to the thread pool
        pool.map(download, self.path_list)
        pool.wait_completion()



from __future__ import with_statement
from __future__ import absolute_import

import re
import sys
import multiprocessing
from boto3.exceptions import S3UploadFailedError
from boto3.s3.transfer import S3Transfer, TransferConfig
from botocore.client import Config
from botocore.exceptions import ClientError

from NDATools.S3Authentication import S3Authentication

if sys.version_info[0] < 3:
    import Queue as queue
    input = raw_input
else:
    import queue
from tqdm import tqdm
from NDATools.Configuration import *
from NDATools.DataManager import *
from NDATools.MultiPartUploads import *
from NDATools.TokenGenerator import *


class Status:
    UPLOADING = 'Uploading'
    SYSERROR = 'SystemError'
    COMPLETE = 'Complete'
    ERROR = 'Error'
    PROCESSING = 'In Progress'
    READY = 'Ready'


class Submission:
    def __init__(self, id, full_file_path, config, resume=False, allow_exit=False, username=None, password=None,
                 thread_num=None, batch_size=None):
        self.config = config
        self.api = self.config.submission_api
        if username:
            self.config.username = username
        if password:
            self.config.password = password
        self.username = self.config.username
        self.password = self.config.password
        self.credentials = S3Authentication(config)
        self.full_file_path = full_file_path
        self.total_upload_size = 0
        self.upload_queue = queue.Queue()
        self.progress_queue = queue.Queue()
        self.thread_num = max([1, multiprocessing.cpu_count() - 1])
        if thread_num:
            self.thread_num = thread_num
        self.batch_size = 10000
        if batch_size:
            self.batch_size = batch_size
        self.batch_status_update = []
        self.directory_list = self.config.directory_list
        self.credentials_list = []
        self.associated_files = []
        self.status = None
        self.total_files = None
        self.total_progress = None
        self.upload_tries = 0
        self.max_submit_time = 120
        self.url = self.config.datamanager_api
        self.resume = resume
        if self.resume:
            self.submission_id = id
            self.no_match = []
            self.no_read_access = set()
        else:
            self.package_id = id
        self.exit = allow_exit
        self.all_mpus = []

    def submit(self):
        response, session = api_request(self, "POST", "/".join([self.api, self.package_id]))
        if response:
            self.status = response['submission_status']
            self.submission_id = response['submission_id']
        else:
            message = 'There was an error creating your submission'
            if self.exit:
                exit_client(signal=signal.SIGTERM,
                        message=message)
            else:
                raise Exception("{}\n{}".format('SubmissionError', message))

    def check_status(self):
        response, session = api_request(self, "GET", "/".join([self.api, self.submission_id]))

        if response:
            self.status = response['submission_status']
        else:
            message='An error occurred while checking submission {} status.'.format(self.submission_id)
            if self.exit:
                exit_client(signal=signal.SIGTERM,
                            message=message)
            else:
                raise Exception("{}\n{}".format('StatusError', message))

    def create_file_info_list(self, response):

        return [{'submissionFileId': file['id'], 'destination_uri': file['file_remote_path']} for file in response if
                file['status'] != Status.COMPLETE]

    def get_multipart_credentials(self, file_ids):
        all_credentials = []
        batched_ids = [file_ids[i:i + self.batch_size] for i in range(0, len(file_ids), self.batch_size)]

        for ids in batched_ids:
            credentials_list, session = api_request(self, "POST", "/".join(
                [self.api, self.submission_id, 'files/batchMultipartUploadCredentials']), data=json.dumps(ids))
            all_credentials = all_credentials + credentials_list['credentials']
            time.sleep(2)

        return all_credentials

    def generate_data_for_request(self, status, id_list=[]):
        batch_status_update = []

        if self.credentials_list:
            id_list = self.credentials_list

        for cred in id_list:
            file = cred['destination_uri'].split('/')
            file = '/'.join(file[4:])
            size = self.full_file_path[file][1]
            update = {
                "id": str(cred['submissionFileId']),
                "md5sum": "None",
                "size": size,
                "status": status
            }
            batch_status_update.append(update)

        self.batch_status_update = batch_status_update
        return batch_status_update

    @property
    def incomplete_files(self):

        response, session = api_request(self, "GET", "/".join([self.api, self.submission_id, 'files']))
        self.full_file_path = {}
        self.associated_files = []
        if response:
            for file in response:
                if file['status'] != Status.COMPLETE:
                    self.associated_files.append(file['file_user_path'])
            return len(self.associated_files) > 0
        else:
            message='There was an error requesting files for submission {}.'.format(self.submission_id)
            if self.exit:
                exit_client(signal=signal.SIGTERM,
                            message=message)
            else:
                raise Exception("{}\n{}".format('SubmissionError', message))

    def check_submitted_files(self):
        response, session = api_request(self, "GET", "/".join([self.api, self.submission_id, 'files']))

        file_info = self.create_file_info_list(response)

        data = self.generate_data_for_request(Status.COMPLETE, file_info)
        url = "/".join([self.api, self.submission_id, 'files/batchUpdate'])
        auth = requests.auth.HTTPBasicAuth(self.config.username, self.config.password)
        headers = {'content-type': 'application/json'}

        batched_data = [data[i:i + self.batch_size] for i in
                        range(0, len(data),
                              self.batch_size)]
        unsubmitted_ids = []

        for batch in batched_data:
            session = requests.session()
            r = session.send(requests.Request('PUT', url, headers, auth=auth, data=json.dumps(batch)).prepare(),
                             timeout=300, stream=False)
            response = json.loads(r.text)
            errors = response['errors']
            for e in errors:
                unsubmitted_ids.append(e['submissionFileId'])

        file_ids = [int(ids['submissionFileId']) for ids in file_info]
        new_ids = set(file_ids) & set(unsubmitted_ids)

        self.credentials_list = self.get_multipart_credentials(list(new_ids))

        self.update_full_file_paths()

    def update_full_file_paths(self):
        full_file_path = {}

        for credentials in self.credentials_list:
            full_path = (credentials['destination_uri'].split(self.submission_id)[1][1:])
            for key,value in self.full_file_path.items():
                if full_path == key:
                    full_file_path[key] = value
                    break
        self.full_file_path = full_file_path

    def recollect_file_search_info(self):
        retry = input('Press the "Enter" key to specify directory/directories OR an s3 location by entering -s3 '
                      '<bucket name> to locate your associated files:')
        response = retry.split(' ')
        self.directory_list = response
        if response[0] == '-s3':
            self.source_bucket = response[1]
            self.source_prefix = input('Enter any prefix for your S3 object, or hit "Enter": ')
            if self.source_prefix == "":
                self.source_prefix = None


    def found_all_files(self, directories=None, source_bucket=None, source_prefix=None, retry_allowed=False):
        def raise_error(error, l = []):
            m = '\n'.join([error] + list(set(l)))
            raise Exception(m)

        self.directory_list = directories
        self.source_bucket = source_bucket
        self.source_prefix = source_prefix

        if not self.directory_list and not self.source_bucket:
            if retry_allowed:
                self.recollect_file_search_info()
            else:
                error = 'Missing directory and/or an S3 bucket.'
                raise_error(error)

        if not self.no_match:
            for file in self.associated_files:
                self.no_match.append(file)

        # local files
        if self.directory_list:
            parse_local_files(self.directory_list, self.no_match, self.full_file_path, self.no_read_access,
                              self.config.skip_local_file_check)

        # files in s3
        no_access_buckets = []
        if self.source_bucket:
            s3_client = self.credentials.get_s3_client()
            for file in self.no_match[:]:
                key = file
                if self.source_prefix:
                    key = '/'.join([self.source_prefix, file])
                file_name = '/'.join(['s3:/', self.source_bucket, key])
                try:
                    response = s3_client.head_object(Bucket=self.source_bucket, Key=key)
                    self.full_file_path[file] = (file_name, int(response['ContentLength']))
                    self.no_match.remove(file)
                except botocore.exceptions.ClientError as e:
                    # If a client error is thrown, then check that it was a 404 error.
                    # If it was a 404 error, then the bucket does not exist.
                    error_code = int(e.response['Error']['Code'])
                    if error_code == 404:
                        pass
                    if error_code == 403:
                        no_access_buckets.append(self.source_bucket)
                        pass

        if self.no_match:
            if no_access_buckets:
                message = 'Your user does NOT have access to the following buckets. Please review the bucket ' \
                          'and/or your AWS credentials and try again.'
                if retry_allowed:
                    print('\n', message)
                    for b in no_access_buckets:
                        print(b)
                else:
                    error = "".join(['Bucket Access:', message])
                    raise_error(error, no_access_buckets)
            message = 'You must make sure all associated files listed in your validation file' \
                      ' are located in the specified directory or AWS bucket. Associated file not found in specified directory:\n'
            if retry_allowed:
                print('\n', message)
                for file in self.no_match:
                    print(file)
                self.recollect_file_search_info()
                self.found_all_files(self.directory_list, self.source_bucket, self.source_prefix, retry_allowed=True)
            else:
                error = "".join(['Missing Files:', message])
                raise_error(error, self.no_match)

        while self.no_read_access:
            message = 'You must make sure you have read-access to all the of the associated files listed in your validation file. ' \
                      'Please update your permissions for the following associated files:\n'
            if retry_allowed:
                print(message)
                for file in self.no_read_access:
                    print(file)
                self.recollect_file_search_info()
                [self.no_read_access.remove(i) for i in
                    [file for file in self.no_read_access if check_read_permissions(file)]]
            else:
                error = "".join(['Read Permission Error:', message])
                raise_error(error, self.no_match)

        self.config.directory_list = self.directory_list
        self.config.source_bucket = self.source_bucket
        self.config.source_prefix = self.source_prefix

        try:
            return len(self.directory_list) > 0
        except TypeError:
            return len(self.source_bucket) > 0

    def batch_update_status(self, status=Status.COMPLETE):
        list_data = self.generate_data_for_request(status)
        url = "/".join([self.api, self.submission_id, 'files/batchUpdate'])
        data_to_dump = [list_data[i:i + self.batch_size] for i in range(0, len(list_data), self.batch_size)]
        for d in data_to_dump:
            data = json.dumps(d)
            api_request(self, "PUT", url, data=data)

    def complete_partial_uploads(self):

        bucket = (self.credentials_list[0]['destination_uri']).split('/')[2]
        prefix = 'submission_{}'.format(self.submission_id)

        # use self.credentials list to pass in a set of temporary tokens, which should allow listMultiPartUpload
        # for the entire submission and not just the specific file object.
        access_key = self.credentials_list[0]['access_key']
        secret_key = self.credentials_list[0]['secret_key']
        session_token = self.credentials_list[0]['session_token']

        multipart_uploads = MultiPartsUpload(bucket, prefix, self.config, access_key, secret_key, session_token)
        multipart_uploads.get_multipart_uploads()

        for upload in multipart_uploads.incomplete_mpu:
            self.all_mpus.append(upload)

    def submission_upload(self, hide_progress):
        with self.upload_queue.mutex:
            self.upload_queue.queue.clear()
        with self.progress_queue.mutex:
            self.progress_queue.queue.clear()
        self.total_files = len(self.full_file_path)
        for associated_file in self.full_file_path:
            path, file_size = self.full_file_path[associated_file]
            self.total_upload_size += file_size

        response, session = api_request(self, "GET", "/".join([self.api, self.submission_id, 'files']))
        if response:
            if not self.resume:
                file_id_info = self.create_file_info_list(response)
                file_ids = [int(ids['submissionFileId']) for ids in file_id_info]
                self.credentials_list = self.get_multipart_credentials(file_ids)
                self.batch_update_status(status=Status.PROCESSING)

            if not hide_progress:
                self.total_progress = tqdm(total=self.total_upload_size,
                                           position=0,
                                           unit_scale=True,
                                           unit="bytes",
                                           desc="Total Upload Progress",
                                           ascii=os.name == 'nt',
                                           dynamic_ncols=True)

            workers = []
            for x in range(self.thread_num):
                worker = Submission.S3Upload(x, self.config, self.upload_queue, self.full_file_path, self.submission_id,
                                             self.progress_queue, self.credentials_list, self.all_mpus)

                workers.append(worker)
                worker.daemon = True
                worker.start()

            for file in response:
                if file['status'] != Status.COMPLETE:
                    self.upload_queue.put([file, False])
            self.upload_queue.put(["STOP", False])
            self.upload_tries += 1
            while any(map(lambda w: w.is_alive(), workers)):
                if not hide_progress:
                    for progress in iter(self.progress_queue.get, None):
                        if (self.total_progress.n < self.total_progress.total
                                and progress <= (self.total_progress.total - self.total_progress.n)):
                            self.total_progress.update(progress)
                time.sleep(0.1)
            if not hide_progress:
                if self.total_progress.n < self.total_progress.total:
                    self.total_progress.update(self.total_progress.total - self.total_progress.n)
                self.total_progress.close()
            session = None

        else:
            message='There was an error requesting submission {}.'.format(self.submission_id)
            if self.exit:
                exit_client(signal=signal.SIGTERM,
                            message=message)
            else:
                raise Exception("{}\n{}".format('SubmissionError', message))

        self.batch_update_status()

        if not hide_progress:
            print('\nUploads complete.')
            print('Checking Submission Status.')
        self.check_status()
        if self.status == Status.UPLOADING:
            if not self.incomplete_files:
                t1 = time.time()
                while self.status == Status.UPLOADING and (time.time() - t1) < self.max_submit_time:
                    self.check_status()
                    timeout_message = 'Timed out while waiting for submission status to change. You may try again by resuming the submission.'
                    if not hide_progress:
                        sys.stdout.write('.')
                if self.status == Status.UPLOADING:
                    if self.exit:
                        exit_client(signal=signal.SIGTERM,
                                    message=timeout_message)
                    else:
                        raise Exception("{}\n{}".format('TimeOutError', timeout_message))

            else:
                print('There was an error transferring some files, trying again')
                if self.found_all_files and self.upload_tries < 5:
                    self.submission_upload(hide_progress)
        if self.status != Status.UPLOADING and hide_progress:
            return

    class S3Upload(threading.Thread):
        def __init__(self, index, config, upload_queue, full_file_path, submission_id, progress_queue, credentials_list,
                     all_mpus):
            threading.Thread.__init__(self)
            self.config = config
            self.upload_queue = upload_queue
            self.upload = None
            self.upload_tries = 0
            self.api = self.config.submission_api
            self.username = self.config.username
            self.password = self.config.password
            self.source_bucket = self.config.source_bucket
            self.source_prefix = self.config.source_prefix
            self.credentials = S3Authentication(self.config)
            self.full_file_path = full_file_path
            self.credentials_list = credentials_list
            self.submission_id = submission_id
            self.all_mpus = all_mpus
            self.index = index + 1
            self.progress_queue = progress_queue
            self.shutdown_flag = threading.Event()
            self.expired = False

        def upload_config(self):
            local_path = sanitize_file_path(self.upload['file_user_path'])
            remote_path = self.upload['file_remote_path']
            file_id = self.upload['id']
            for cred in self.credentials_list:
                if str(cred['submissionFileId']) == file_id:
                    credentials = {'access_key': cred['access_key'], 'secret_key': cred['secret_key'],
                                   'session_token': cred['session_token']}
                    break
            paths = remote_path.split('/')
            bucket = paths[2]
            key = "/".join(paths[3:])
            full_path, file_size = self.full_file_path[local_path]
            return file_id, credentials, bucket, key, full_path, file_size

        def update_tokens(self):
            link = self.upload['_links']['multipartUploadCredentials']['href']
            credentials, session = api_request(self, "GET", link)

            for creds in self.credentials_list:
                if creds['fileId'] == credentials['fileId']:
                    self.credentials_list.remove(creds)
                    break
            self.credentials_list.append(credentials)

            return credentials

        def add_back_to_queue(self, bucket, prefix):
            # If expired token:
            #  1. Add to upload_queue so it is picked up by another thread again
            self.upload_queue.put([self.upload, True])  # need to figure out how to test this and if it is the best way to retry a single file upload
            self.upload_queue.put(["STOP", False])  # need to add the sentinel value again

            # 2. Add a function to get new credentials and update self.credentials_list AND
            # 3. In upload.config, edit it so it knows we need new credentials
            new_credentials = self.update_tokens()

            # 4. Check if file has mpu, add to self.mpu

            multipart_uploads = MultiPartsUpload(bucket, prefix, self.config, new_credentials['access_key'],
                                                 new_credentials['secret_key'], new_credentials['session_token'])
            multipart_uploads.get_multipart_uploads()

            self.all_mpus = multipart_uploads.incomplete_mpu

        class UpdateProgress(object):

            def __init__(self, progress_queue):
                self.progress_queue = progress_queue

            def __call__(self, bytes_amount):
                self.progress_queue.put(bytes_amount)

        def run(self):
            while True and not self.shutdown_flag.is_set():
                if self.upload and self.upload_tries < 5:
                    self.upload_tries += 1
                else:
                    upload = self.upload_queue.get()

                self.upload = upload[0]
                self.expired = upload[1]
                if self.upload == "STOP":
                    if self.upload_queue.qsize() == 0:
                        self.upload_queue.put(["STOP", False])
                        self.shutdown_flag.set()
                        break
                    else:
                        self.progress_queue.put(None)
                        self.upload_tries = 0
                        self.upload = None
                        self.upload_queue.task_done()
                        upload = self.upload_queue.get()
                        self.upload = upload[0]
                        self.expired = upload[1]

                file_id, credentials, bucket, key, full_path, file_size = self.upload_config()
                prefix = 'submission_{}'.format(self.submission_id)


                """
                Methods for  file transfer: 

                If the source file is local, use the credentials supplied by the submission API to upload from local 
                file to remote S3 location.

                If the source file is from S3:
                a) check settings.cfg for permanent user credentials (aws_access_key, aws_secret_key)
                b) if permanent credentials are provided, use them to retrieve the source file,
                c) if not provided use a FederationUser token from DataManager API to retreive the source file,
                d) use credentials supplied by the submission API to upload to remote S3 location.
                
                If the file was uploaded using multi-part, it will first complete the multi part uploads.
                """

                expired_error = False
                mpu_exist = False
                for upload in self.all_mpus:
                    if upload['Key'] == key:
                        mpu_exist = True
                        mpu_to_complete = upload
                        break

                if full_path.startswith('s3'):

                    """
                    Assumes you are uploading from external s3 bucket. SOURCE_BUCKET and SOURCE_PREFIX are hard-coded 
                    values, which specify where the object should be copied from (i.e., 100206 subject directory can be
                    located in s3://hcp-openaccess-temp, with a prefix of HCP_1200).

                    Creates source and destination clients for S3 tranfer. If supplied in settings.cfg uses permanent 
                    credentials for accessing source buckets. If permanent credentials are not supplied in
                    settings.cfg, uses a tempoary FederationUser Token from DataManager API to access source bucket. 

                    The transfer uses a file streaming method by streaming the body of the file into memory and uploads 
                    the stream in chunks using AWS S3Transfer. This Transfer client will automatically use multi part 
                    uploads when necessary. To maximize efficiency, only files greater than 1e8 bytes are uploaded using 
                    the multi part upload. Smaller files are uploaded in one part.  

                    After each successful transfer, the script will change the status of the file to complete in NDA's 
                    submission webservice. 

                    NOTE: For best results and to be cost effective, it is best to perform this file transfer in an AWS 
                    EC2 instance.
                    """

                    tqdm.monitor_interval = 0

                    s3_config = Config(connect_timeout=240, read_timeout=240)
                    self.source_s3 = self.credentials.get_s3_resource(s3_config)

                    source_key = key.split('/')[1:]
                    source_key = '/'.join(source_key)
                    self.source_key = '/'.join([self.source_prefix, source_key])
                    self.fileobj = self.source_s3.Object(self.source_bucket, self.source_key).get()['Body']

                    if mpu_exist:
                        u = UploadMultiParts(mpu_to_complete, self.full_file_path, bucket, prefix, self.config, credentials, file_size)
                        u.get_parts_information()
                        if not self.expired:
                            self.progress_queue.put(u.completed_bytes)
                        seq = 1

                        for buffer in self.fileobj.iter_chunks(chunk_size=u.chunk_size):
                            if seq in u.parts_completed:
                                part = u.parts[seq - 1]
                                u.check_md5(part, buffer)
                            else:
                                try:
                                    u.upload_part(buffer, seq)
                                    self.progress_queue.put(len(buffer))
                                    # upload missing part
                                except Exception as error:
                                    e = str(error)
                                    if "ExpiredToken" in e:
                                        self.add_back_to_queue(bucket, prefix)
                                        expired_error = True
                                    else:
                                        raise error
                            seq += 1
                        if not expired_error:
                            u.complete()
                        self.progress_queue.put(None)

                    else:

                        # set chunk size dynamically to based on file size
                        if file_size > 9999:
                            multipart_chunk_size = (file_size // 9999)
                        else:
                            multipart_chunk_size = file_size
                        transfer_config = TransferConfig(multipart_threshold=100 * 1024 * 1024,
                                                         multipart_chunksize=multipart_chunk_size)

                        self.dest = S3Authentication.get_s3_client_with_config(credentials['access_key'],
                                                                               credentials['secret_key'],
                                                                               credentials['session_token'])
                        self.dest_bucket = bucket
                        self.dest_key = key
                        self.temp_key = self.dest_key + '.temp'

                        try:
                            self.dest.upload_fileobj(
                                self.fileobj,
                                self.dest_bucket,
                                self.dest_key,
                                Callback=self.UpdateProgress(self.progress_queue),
                                Config=transfer_config
                            )
                        except boto3.exceptions.S3UploadFailedError as error:
                            e = str(error)
                            if "ExpiredToken" in e:
                                self.add_back_to_queue(bucket, prefix)
                                self.credentials = S3Authentication(self.config)
                            else:
                                raise error
                    self.progress_queue.put(None)

                else:
                    """
                    Assumes the file is being uploaded from local file system
                    """

                    if mpu_exist:
                        u = UploadMultiParts(mpu_to_complete, self.full_file_path, bucket, prefix, self.config,
                                             credentials, file_size)
                        u.get_parts_information()
                        if not self.expired:
                            self.progress_queue.put(u.completed_bytes)
                        seq = 1

                        with open(full_path, 'rb') as f:
                            while True:
                                buffer_start = u.chunk_size * (seq - 1)
                                f.seek(buffer_start)
                                buffer = f.read(u.chunk_size)
                                if len(buffer) == 0:  # EOF
                                    break
                                if seq in u.parts_completed:
                                    part = u.parts[seq - 1]
                                    u.check_md5(part, buffer)
                                else:
                                    try:
                                        u.upload_part(buffer, seq)
                                        self.progress_queue.put(len(buffer))
                                    except Exception as error:
                                        e = str(error)
                                        if "ExpiredToken" in e:
                                            self.add_back_to_queue(bucket, prefix)
                                            expired_error = True
                                            break
                                        else:
                                            raise error

                                seq += 1
                        if not expired_error:
                            u.complete()
                        self.progress_queue.put(None)
                    else:
                        if credentials:
                            s3 = S3Authentication.get_s3_client_with_config(credentials['access_key'],
                                                                            credentials['secret_key'],
                                                                            credentials['session_token'])
                            config = TransferConfig(
                                multipart_threshold=100 * 1024 * 1024,
                                max_concurrency=2,
                                num_download_attempts=10)

                            s3_transfer = S3Transfer(s3, config)
                            tqdm.monitor_interval = 0
                            try:
                                s3_transfer.upload_file(full_path, bucket, key,
                                                        callback=self.UpdateProgress(self.progress_queue))
                            except boto3.exceptions.S3UploadFailedError as error:
                                e = str(error)
                                if "ExpiredToken" in e:
                                    self.add_back_to_queue(bucket, prefix)
                                    self.credentials = S3Authentication(self.config)
                                else:
                                    raise error

                            self.progress_queue.put(None)

                        else:
                            print('There was an error uploading {} after {} retry attempts'.format(full_path,
                                                                                                   self.upload_tries))
                            continue

                self.upload_tries = 0
                self.upload = None
                self.upload_queue.task_done()



import logging

import NDATools

from NDATools.clientscripts.vtcmd import build_package
from tests import test_validation

from unittest import TestCase


class TestBuildPackage(TestCase):
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s:%(levelname)s:%(message)s")

    def test_build_package_success(self):
        validation_results, config = test_validation.TestValidation.test_manifest_validation_build_package_config(self)
        logging.debug(validation_results)

        uuid = validation_results[0]
        associated_files = validation_results[1]

        package_results = build_package(uuid, associated_files, config=config)
        self.assertIsNotNone(package_results[0])
        self.assertIsNotNone(package_results[1])
        logging.debug(package_results)

        return package_results, validation_results, config

    def test_s3_build_package(self):
        validation_results, config = test_validation.TestValidation.test_s3_validation(self)
        logging.debug(validation_results)

        uuid = validation_results[0]
        associated_files = validation_results[1]

        package_results = build_package(uuid, associated_files, config=config)
        self.assertIsNotNone(package_results[0])
        self.assertIsNotNone(package_results[1])
        logging.debug(package_results)

        return package_results, validation_results, config



from __future__ import with_statement
from __future__ import absolute_import

from NDATools.TokenGenerator import *

class DataManager:

    def __init__(self, config):
        self.config = config
        self.username = self.config.username
        self.password = self.config.password
        self.api = self.config.datamanager_api
        self.credentials = self.get_data_manager_credentials

    @property
    def get_data_manager_credentials(self):
        data_manager_token = NDATokenGenerator(self.api).generate_token(self.username,
                                                                        self.password)
        credentials = {'aws_access_key_id': data_manager_token.access_key,
                       'aws_secret_access_key': data_manager_token.secret_key,
                       'aws_session_token': data_manager_token.session}
        return credentials